#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <string.h>

#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_nvic.h"
#include "nrf_log.h"
#include "MegaSig_BLE_Function.h"
#include "MegaSig_Global.h"

void ble_usb_printf()
{
  usb_printf_len((char *)op_bfr, op_bfr_count);
  memset(op_bfr, 0, BLE_MAX_BFR_SIZE);
  op_bfr_count = 0;
}

void ble_usb_add_array(const uint8_t *value, uint32_t len)
{
  for (int i = 0; i < len; i++)
  {
    if (op_bfr_count + 2 >= BLE_MAX_BFR_SIZE)
      ble_usb_printf();
    int count = sprintf((char *)op_bfr + op_bfr_count, "%02x", value[i]);
    op_bfr_count += count;
  }
}

void ble_usb_add_value_uint8(uint8_t value, uint8_t type)
{
  if (op_bfr_count + 3 >= BLE_MAX_BFR_SIZE)
    ble_usb_printf();
  switch (type)
  {
  case 0:
  {
    int count = snprintf((char *)op_bfr + op_bfr_count, BLE_MAX_BFR_SIZE - op_bfr_count, "%d", value);
    op_bfr_count += count;
    break;
  }

  case 1:
  {
    int count = sprintf((char *)op_bfr + op_bfr_count, "%02x", value);
    op_bfr_count += count;
    break;
  }
  }
}

void ble_usb_add_value_uint16(uint16_t value, uint8_t type)
{
  if (op_bfr_count + 5 >= BLE_MAX_BFR_SIZE)
    ble_usb_printf();
  switch (type)
  {
  case 0:
  {
    uint16_t count = snprintf((char *)op_bfr + op_bfr_count, BLE_MAX_BFR_SIZE - op_bfr_count, "%u", value);
    op_bfr_count += count;
    break;
  }

  case 1:
  {
    uint16_t count = snprintf((char *)op_bfr + op_bfr_count, BLE_MAX_BFR_SIZE - op_bfr_count, "%04x", value);
    op_bfr_count += count;
    break;
  }
  }
}

void ble_usb_add_value_uint32(uint32_t value, uint8_t type)
{
  if (op_bfr_count + 10 >= BLE_MAX_BFR_SIZE)
    ble_usb_printf();
  switch (type)
  {
  case 0:
  {
    uint16_t count = snprintf((char *)op_bfr + op_bfr_count, BLE_MAX_BFR_SIZE - op_bfr_count, "%u", value);
    op_bfr_count += count;
    break;
  }

  case 1:
  {
    uint16_t count = snprintf((char *)op_bfr + op_bfr_count, BLE_MAX_BFR_SIZE - op_bfr_count, "%08x", value);
    op_bfr_count += count;
    break;
  }
  }
}

void ble_usb_add_string(uint8_t *string, uint16_t len)
{
  if (op_bfr_count + len >= BLE_MAX_BFR_SIZE)
    ble_usb_printf();
  memcpy(op_bfr + op_bfr_count, string, len);
  op_bfr_count += len;
}

void ble_usb_add_char_props(ble_gatt_char_props_t props)
{
  if (props.broadcast)
    ble_usb_add_string((uint8_t *)"broadcast, ", 9);
  if (props.read)
    ble_usb_add_string((uint8_t *)"read, ", 4);
  if (props.write_wo_resp)
    ble_usb_add_string((uint8_t *)"write_wo_resp, ", 13);
  if (props.write)
    ble_usb_add_string((uint8_t *)"write, ", 6);
  if (props.notify)
    ble_usb_add_string((uint8_t *)"notify, ", 7);
  if (props.indicate)
    ble_usb_add_string((uint8_t *)"indicate, ", 9);
  if (props.auth_signed_wr)
    ble_usb_add_string((uint8_t *)"auth_signed_wr, ", 14);

  memset(op_bfr + op_bfr_count - 2, 0, 3);
  op_bfr_count -= 2;
}

uint8_t wait_trigger_evt(uint16_t target_evt)
{
  uint32_t start_time = 0;
  uint32_t c_time = 0;
  start_time = app_timer_cnt_get();
  start_time = APP_TIMER_TICKS_TO_MS(start_time);
  while (((c_time - start_time) <= 10000) && (trigger_evt_id != target_evt))
  {
    nrf_delay_us(10);
    c_time = app_timer_cnt_get();
    c_time = APP_TIMER_TICKS_TO_MS(c_time);
  }
  if ((c_time - start_time) >= 10000)
  {
    return NRF_ERROR_TIMEOUT;
  }
  return 0;
}

/**@brief Function for converting a BLE address to a string.
 *
 * @param[in] address       Bluetooth Low Energy address.
 * @param[out] string_buffer The serial port the target nRF5 device is connected to.
 */
void ble_address_to_string_convert(ble_gap_addr_t address, uint8_t *string_buffer)
{
  const int address_length = 6;
  char temp_str[3];
  current_scan_info.peer_addr.addr_type = address.addr_type;
  for (int i = address_length - 1; i >= 0; --i)
  {
    strcat((char *)string_buffer, temp_str);
    current_scan_info.peer_addr.addr[i] = address.addr[i];
  }
}

/**
 * @brief Parses advertisement data, providing length and location of the field in case
 *        matching data is found.
 *
 * @param[in]  Type of data to be looked for in advertisement data.
 * @param[in]  Advertisement report length and pointer to report.
 * @param[out] If data type requested is found in the data report, type data length and
 *             pointer to data will be populated here.
 *
 * @retval NRF_SUCCESS if the data type is found in the report.
 * @retval NRF_ERROR_NOT_FOUND if the data type could not be found.
 */
uint32_t adv_report_parse(uint8_t type, data_t *p_advdata, data_t *p_typedata)
{
  uint32_t index = 0;
  uint8_t *p_data;

  p_data = p_advdata->p_data;

  while (index < p_advdata->data_len)
  {
    uint8_t field_length = p_data[index];
    uint8_t field_type = p_data[index + 1];

    if (field_type == type)
    {
      p_typedata->p_data = &p_data[index + 2];
      p_typedata->data_len = field_length - 1;
      return NRF_SUCCESS;
    }
    index += field_length + 1;
  }
  return NRF_ERROR_NOT_FOUND;
}

/**@brief Function for searching a given name in the advertisement packets.
 *
 * @details Use this function to parse received advertising data and to find a given
 * name in them either as 'complete_local_name' or as 'short_local_name'.
 *
 * @param[in]   p_adv_report   advertising data to parse.
 * @param[in]   name_to_find   name to search.
 * @return   true if the given name was found, false otherwise.
 */
bool find_adv_name(const ble_gap_evt_adv_report_t *p_adv_report, const char *name_to_find)
{
  uint32_t error_code;
  data_t adv_data;
  data_t dev_name;

  // Initialize advertisement report for parsing
  adv_data.p_data = (uint8_t *)p_adv_report->data.p_data;
  adv_data.data_len = p_adv_report->data.len;

  // search for advertising names
  error_code = adv_report_parse(BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME, &adv_data, &dev_name);
  if (error_code == NRF_SUCCESS)
  {
    if (memcmp(name_to_find, dev_name.p_data, dev_name.data_len) == 0)
    {
      return true;
    }
  }
  else
  {
    // Look for the short local name if it was not found as complete
    error_code = adv_report_parse(BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME, &adv_data, &dev_name);
    if (error_code != NRF_SUCCESS)
    {
      return false;
    }
    if (memcmp(name_to_find, dev_name.p_data, dev_name.data_len) == 0)
    {
      return true;
    }
  }
  return false;
}

/**@brief Function for setting configuration for the BLE stack.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t ble_cfg_set(uint8_t conn_cfg_tag, uint32_t ram_start)
{
  uint32_t error_code;
  ble_cfg_t ble_cfg;

  // Configure the connection roles.
  memset(&ble_cfg, 0, sizeof(ble_cfg));

  ble_cfg.gap_cfg.role_count_cfg.adv_set_count = BLE_GAP_ADV_SET_COUNT_DEFAULT;
  ble_cfg.gap_cfg.role_count_cfg.periph_role_count = 1;
  ble_cfg.gap_cfg.role_count_cfg.central_role_count = 1;
  ble_cfg.gap_cfg.role_count_cfg.central_sec_count = 1;

  error_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
  if (error_code != NRF_SUCCESS)
  {
    // //printf("sd_ble_cfg_set() failed when attempting to set BLE_GAP_CFG_ROLE_COUNT. Error code: ""0x%02X\n",error_code);
    BLE_USB_PRINT_MSG_NUM("sd_ble_cfg_set() failed when attempting to set BLE_GAP_CFG_ROLE_COUNT. Error code: 0x ", error_code, 1);
    return error_code;
  }

  memset(&ble_cfg, 0x00, sizeof(ble_cfg));
  ble_cfg.conn_cfg.conn_cfg_tag = conn_cfg_tag;
  ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = NRF_SDH_BLE_GATT_MAX_MTU_SIZE;

  error_code = sd_ble_cfg_set(BLE_CONN_CFG_GATT, &ble_cfg, ram_start);
  if (error_code != NRF_SUCCESS)
  {
    // //printf("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GATT. Error code: ""0x%02X\n",error_code);
    BLE_USB_PRINT_MSG_NUM("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GATT. Error code: 0x", error_code, 1);
    return error_code;
  }

  memset(&ble_cfg, 0, sizeof(ble_cfg));
  ble_cfg.conn_cfg.conn_cfg_tag = conn_cfg_tag;
  ble_cfg.conn_cfg.params.gap_conn_cfg.conn_count = 1;
  ble_cfg.conn_cfg.params.gap_conn_cfg.event_length = NRF_SDH_BLE_GAP_EVENT_LENGTH;

  error_code = sd_ble_cfg_set(BLE_CONN_CFG_GAP, &ble_cfg, ram_start);
  if (error_code != NRF_SUCCESS)
  {
    // //printf("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GATT. Error code: ""0x%02X\n",error_code);
    BLE_USB_PRINT_MSG_NUM("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GATT. Error code: 0x", error_code, 1);
    return error_code;
  }

  return NRF_SUCCESS;
}

/**@brief Function for initializing the BLE stack.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_stack_init()
{
  uint32_t error_code;

  uint32_t ram_start = 0;
  error_code = nrf_sdh_enable_request();
  BLE_USB_PRINT_MSG_NUM("NRF SDH Enabel Request Error Code = 0x", error_code, 1);
  error_code = app_timer_init();
  BLE_USB_PRINT_MSG_NUM("APP TIMER Init Error Code = 0x", error_code, 1);
  // error_code = nrf_sdh_ble_default_cfg_set(1, &ram_start);
  error_code = nrf_sdh_ble_app_ram_start_get(&ram_start);
  BLE_USB_PRINT_MSG_NUM("NRF SDH APP Ram Start Get Error Code = 0x", error_code, 1);
  error_code = ble_cfg_set(m_config_id, ram_start);
  BLE_USB_PRINT_MSG_NUM("BLE CFG Set Error Code = 0x", error_code, 1);
  error_code = nrf_sdh_ble_enable(&ram_start);
  BLE_USB_PRINT_MSG_NUM("BLE Enable Error Code = 0x", error_code, 1);

  return error_code;
}

/**@brief Start scanning (GAP Discovery procedure, Observer Procedure).
 * *
 * @return NRF_SUCCESS on successfully initiating scanning procedure, otherwise an error code.
 */
uint32_t scan_start()
{
  m_adv_report_buffer.p_data = mp_data;
  m_adv_report_buffer.len = sizeof(mp_data);

  uint32_t error_code = sd_ble_gap_scan_start(&m_scan_param, &m_adv_report_buffer);

  if (error_code != NRF_SUCCESS)
  {
    // //printf("Scan start failed with error code: %d\n", error_code);
    BLE_USB_PRINT_MSG_NUM("GAP Scan Start Error Code: 0x", error_code, 1);
  }
  else
  {
    // //printf("Scan started\n");
    ble_usb_add_string("Scan Started\n", 13);
    ble_usb_printf();
  }

  return error_code;
}

uint32_t gattc_read(uint16_t target_handle)
{
  uint32_t error_code;

  // Use sd_ble_gattc_read can read service uuid.
  error_code = sd_ble_gattc_read(m_connection_handle, target_handle, 0);
  if (error_code != NRF_SUCCESS)
    BLE_USB_PRINT_MSG_NUM("Gattc Read Error Code: 0x", error_code, 1);
  return error_code;
}

void reverseArray(uint8_t arr[], uint8_t start, uint8_t end)
{
  while (start < end)
  {
    // Ԫ
    int temp = arr[start];
    arr[start] = arr[end];
    arr[end] = temp;

    // ƶָ
    start++;
    end--;
  }
}

uint32_t _128bit_uuid_service_discover()
{
  uint32_t error_code;

  uint16_t size = service_128_handle_index;
  if (size > 0)
  {
    for (int index = 0; index < size; index++)
    {
      trigger_evt_id = 0;
      // Use sd_ble_gattc_read can read service uuid.
      error_code = sd_ble_gattc_read(m_connection_handle,
                                     gattc_128_service_handle[index].handle, 0);
      if (error_code != NRF_SUCCESS)
      {
        BLE_USB_PRINT_MSG_NUM("128bit UUID Service Discover Error Code: 0x", error_code, 1);
        return error_code;
      }
      error_code = wait_trigger_evt(BLE_GATTC_EVT_READ_RSP);
      if (error_code != NRF_SUCCESS)
      {
        BLE_USB_PRINT_MSG_NUM("128bit UUID Service Discover Error Code: 0x", error_code, 1);
        return error_code;
      }
    }
  }
  return 0;
}

uint32_t service_discovery_start(uint16_t start_handle, uint8_t *uuid)
{
  uint32_t error_code = 0;

  ble_uuid128_t base_uuid = NUS_BASE_UUID;

  // printf("Discovering primary services\n");

  // Initiate procedure to find the primary BLE_UUID_NUS_SERVICE.

  uint16_t current_size = 1;
  uint16_t _start_handle = start_handle;

  ble_uuid_t target_uuid;
  ble_uuid128_t _128_uuid;

  if (uuid != NULL)
  {
    memcpy(_128_uuid.uuid128, uuid, (sizeof(uint8_t) * 16));
    reverseArray(_128_uuid.uuid128, 0, 16);
    error_code = sd_ble_uuid_vs_add(&_128_uuid, &target_uuid.type);
    if (error_code != NRF_SUCCESS)
      return error_code;
    target_uuid.uuid = (_128_uuid.uuid128[13] << 8) | _128_uuid.uuid128[12];
  }

  while (service_handle_index != current_size)
  {
    current_size = service_handle_index;
    trigger_evt_id = 0;
    if (uuid != NULL)
    {
      error_code = sd_ble_gattc_primary_services_discover(m_connection_handle,
                                                          _start_handle, &target_uuid);
    }
    else
    {
      error_code = sd_ble_gattc_primary_services_discover(m_connection_handle,
                                                          _start_handle, NULL);
    }
    if (error_code != NRF_SUCCESS)
    {
      if (error_code == 0x11)
      {
        error_code = sd_ble_gattc_primary_services_discover(m_connection_handle, _start_handle, NULL);
        if (error_code != NRF_SUCCESS)
        {
          BLE_USB_PRINT_MSG_NUM("Service Discover Error Code: 0x", error_code, 1);
          return error_code;
        }
      }
    }
    error_code = wait_trigger_evt(BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP);
    if (error_code != NRF_SUCCESS)
    {
      BLE_USB_PRINT_MSG_NUM("Service Discover Error Code: 0x", error_code, 1);
      return error_code;
    }

    // Current last handle +1 and then continue the search based on this handle
    _start_handle = gattc_service_handle[service_handle_index].end_handle + 1;

    // When no find new service, service discover ending
    if (service_handle_index == current_size)
      break;
  }
  // //printf("Service discover end\n");
  BLE_USB_PRINT_MSG("Service discover end\n");
  return _128bit_uuid_service_discover();
}

uint32_t gattc_write(uint16_t handle, uint8_t write_op, uint32_t data_len, uint8_t *data)
{
  uint32_t error_code;
  ble_gattc_write_params_t write_params;

  write_params.handle = handle;
  write_params.write_op = write_op;
  write_params.offset = 0;
  write_params.len = data_len;
  write_params.p_value = data;

  error_code = sd_ble_gattc_write(m_connection_handle, &write_params);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Write Error Code: 0x", error_code, 1);
    return error_code;
  }

  return error_code;
}

uint32_t gattc_request_mtu_size(uint8_t mtu_size)
{
  uint32_t error_code;

  sd_ble_gattc_exchange_mtu_request(m_connection_handle, mtu_size);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC MTU Request Error Code: 0x", error_code, 1);
    return error_code;
  }
  return error_code;
}

uint32_t gattc_services_discover(uint16_t start_handle, uint16_t uuid, uint8_t type)
{
  uint32_t error_code;
  ble_uuid_t t_uuid = {uuid, type};

  error_code = sd_ble_gattc_primary_services_discover(m_connection_handle, start_handle, &t_uuid);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Services Discover Error Code: 0x", error_code, 1);
    return error_code;
  }

  return error_code;
}

uint32_t gattc_characteristics_discover(uint16_t start_handle, uint16_t end_handle)
{
  uint32_t error_code;
  ble_gattc_handle_range_t handle_range = {start_handle, end_handle};

  error_code = sd_ble_gattc_characteristics_discover(m_connection_handle, &handle_range);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Characteristics Discover Error Code: 0x", error_code, 1);
    return error_code;
  }

  return error_code;
}

uint32_t gattc_descriptors_discover(uint16_t start_handle, uint16_t end_handle)
{
  uint32_t error_code;
  ble_gattc_handle_range_t handle_range = {start_handle, end_handle};

  error_code = sd_ble_gattc_descriptors_discover(m_connection_handle, &handle_range);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Descriptors Discover Error Code: 0x", error_code, 1);
    return error_code;
  }
  return error_code;
}

uint32_t gattc_attr_info_discover(uint16_t start_handle, uint16_t end_handle)
{
  uint32_t error_code;
  ble_gattc_handle_range_t handle_range = {start_handle, end_handle};

  error_code = sd_ble_gattc_attr_info_discover(m_connection_handle, &handle_range);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Attr Info Discover Error Code: 0x", error_code, 1);
    return error_code;
  }
  return error_code;
}

bool string_to_ble_address_convert(const char *string_buffer, ble_gap_addr_t *p_address)
{
  int addr_bytes[6];

  if (sscanf((const char *)string_buffer, "%02x%02x%02x%02x%02x%02x", &addr_bytes[5],
             &addr_bytes[4], &addr_bytes[3], &addr_bytes[2], &addr_bytes[1], &addr_bytes[0]) != 6)
  {
    return false; // Failed to parse
  }

  for (int i = 0; i < 6; i++)
  {
    p_address->addr[i] = (uint8_t)addr_bytes[i];
  }

  // The original function didn't provide the address type, so this will need to be set based on
  // context or additional info. Here's an example setting it to public:
  p_address->addr_type = BLE_GAP_ADDR_TYPE_RANDOM_PRIVATE_RESOLVABLE;

  return true;
}

void form_bool_array_to_char_props(uint8_t *boolarray, ble_gatts_char_md_t *p_char_md)
{
  p_char_md->char_props.broadcast = boolarray[0];
  p_char_md->char_props.read = boolarray[1];
  p_char_md->char_props.write_wo_resp = boolarray[2];
  p_char_md->char_props.write = boolarray[3];
  p_char_md->char_props.notify = boolarray[4];
  p_char_md->char_props.indicate = boolarray[5];
  p_char_md->char_props.auth_signed_wr = boolarray[6];
}

void uint8_t_array_copy(uint8_t *source, uint8_t *dec, uint16_t len)
{
  for (uint16_t i = 0; i < len; i++)
  {
    dec[i] = source[i];
  }
}

uint32_t b_characteristic_init(uint16_t service_handle, uint16_t char_uuid, uint8_t *char_name,
                               uint8_t char_name_size, uint8_t *char_props, uint8_t read_perm,
                               uint8_t write_perm, uint8_t rd_auth, uint8_t wr_auth,
                               uint8_t ex_reliable_wr, uint8_t ex_wr_aux, uint16_t *handle,
                               uint16_t *cccd_handle, uint16_t *sccd_handle,
                               uint16_t *user_desc_handle)
{
  uint32_t error_code;
  ble_gatts_char_md_t char_md;
  ble_gatts_attr_t attr_char_value;
  ble_uuid_t ble_uuid;
  ble_gatts_attr_md_t attr_md;
  uint16_t attr_char_value_init_len;
  ble_gatts_attr_md_t cccd_md;

  memset(&char_md, 0, sizeof(char_md));

  form_bool_array_to_char_props(char_props, &char_md);
  if (char_name_size != 0)
  {
    ble_gatts_attr_md_t user_desc_md;
    char_md.p_char_user_desc = char_name; // char name
    char_md.char_user_desc_size = char_name_size;
    char_md.char_user_desc_max_size = char_name_size;

    user_desc_md.rd_auth = 0;
    user_desc_md.wr_auth = 0;
    user_desc_md.vloc = BLE_GATTS_VLOC_STACK;
    user_desc_md.vlen = 0;
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&user_desc_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&user_desc_md.write_perm);
    char_md.p_user_desc_md = &user_desc_md;
  }
  else
  {
    char_md.p_user_desc_md = NULL;
  }
  char_md.p_char_pf = NULL; // set char descriptor format
  if (char_md.char_props.notify == 1)
  {
    memset(&cccd_md, 0, sizeof(cccd_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);
    cccd_md.vloc = BLE_GATTS_VLOC_STACK; // Set this service stack save in BLE_STACK
    char_md.p_cccd_md = &cccd_md;        // set cahr cccd;
  }
  char_md.p_sccd_md = NULL;
  char_md.char_ext_props.reliable_wr = ex_reliable_wr;
  char_md.char_ext_props.wr_aux = ex_wr_aux;

  BLE_UUID_BLE_ASSIGN(ble_uuid, char_uuid);

  memset(&attr_md, 0, sizeof(attr_md));

  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
  attr_md.vloc = BLE_GATTS_VLOC_STACK;
  attr_md.rd_auth = rd_auth;
  attr_md.wr_auth = wr_auth;
  attr_md.vlen = 0;

  memset(&attr_char_value, 0, sizeof(attr_char_value));

  attr_char_value_init_len = 0;

  attr_char_value.p_uuid = &ble_uuid;
  attr_char_value.p_attr_md = &attr_md;
  attr_char_value.init_len = attr_char_value_init_len;
  attr_char_value.init_offs = 0;
  attr_char_value.max_len = 20;
  attr_char_value.p_value = NULL;

  ble_gatts_char_handles_t temp_var;
  error_code = sd_ble_gatts_characteristic_add(service_handle, &char_md, &attr_char_value, &temp_var);

  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTS Characteristics Add Error Code: 0x", error_code, 1);
    return error_code;
  }

  *handle = temp_var.value_handle;
  *cccd_handle = temp_var.cccd_handle;
  *sccd_handle = temp_var.sccd_handle;
  *user_desc_handle = temp_var.user_desc_handle;

  // //printf("Characteristics initiated\n");
  BLE_USB_PRINT_MSG("Characteristics Add OK\n");
  return NRF_SUCCESS;
}

uint32_t b_services_init(uint16_t u_uuid)
{
  uint32_t error_code;
  ble_uuid_t ble_uuid;
  uint16_t t_handle;

  BLE_UUID_BLE_ASSIGN(ble_uuid, u_uuid);
  error_code =
      sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid,
                               &t_handle);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTS Service Add Error Code: 0x", error_code, 1);
    return error_code;
  }

  // //printf("Services initiated\n");
  BLE_USB_PRINT_MSG("Service Add OK\n");

  return NRF_SUCCESS;
}

uint32_t advertising_init(uint8_t adv_AD_type, uint8_t *adv_data_data,
                          uint8_t adv_data_len, uint16_t interval, uint16_t duration) // Set adv data
{
  uint32_t error_code = 0;
  uint8_t data_buffer[30];
  uint8_t index = 0;
  const uint8_t *sr_data = NULL;
  const uint8_t sr_data_len = 0;
  ble_gap_adv_data_t adv_data;
  ble_gap_adv_params_t adv_params;
  uint8_t adv_handle;

  data_buffer[index++] = adv_data_len + 1;
  data_buffer[index++] = adv_AD_type;
  memcpy((char *)&data_buffer[index], adv_data_data, adv_data_len);
  index += adv_data_len;
  // printf("data len = 0x%02X\n", adv_data_len);

  memset(&adv_params, 0, sizeof(adv_params));
  adv_params.properties.type = BLE_GAP_ADV_TYPE_CONNECTABLE_SCANNABLE_UNDIRECTED;
  adv_params.p_peer_addr = NULL; // 㲥
  adv_params.duration = duration;
  adv_params.interval = interval / 0.625; // unit 0.625us/cycle 100ms = 100/0.625 = 160cycle

  // ù㲥
  adv_data.adv_data.p_data = adv_data_data;
  adv_data.adv_data.len = adv_data_len;

  // ɨӦݣҪ
  adv_data.scan_rsp_data.p_data = NULL;
  adv_data.scan_rsp_data.len = 0;

  error_code = sd_ble_gap_adv_set_configure(&adv_handle, &adv_data, &adv_params);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GAP Advertising Config Error Code: 0x", error_code, 1);
    return error_code;
  }

  // printf("Advertising data set\n");
  BLE_USB_PRINT_MSG("Advertising OK\n");
  // print adv_handle
  BLE_USB_PRINT_MSG_NUM("Advertising Handle: 0x", adv_handle, 1);
  return NRF_SUCCESS;
}

uint32_t advertising_start(uint16_t *adv_handle, uint16_t count)
{
  uint32_t error_code = 0;

  for (uint16_t i = 0; i < count; i++)
  {
    error_code = sd_ble_gap_adv_start(adv_handle[i], BLE_CONN_CFG_TAG_DEFAULT);
    if (error_code != NRF_SUCCESS)
    {
      BLE_USB_PRINT_MSG_NUM("GAP Advertising Start Error Code: 0x", error_code, 1);
      return error_code;
    }
  }

  BLE_USB_PRINT_MSG("Advertising Start OK\n");
  return NRF_SUCCESS;
}

uint32_t advertising_stop(uint16_t adv_handle)
{
  uint32_t error_code = 0;
  error_code = sd_ble_gap_adv_stop(adv_handle);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GAP Advertising Stop Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("Advertising Stop OK\n");
  return error_code;
}

uint32_t gatts_char_value_set(uint16_t handle, uint8_t *data, uint16_t data_size, uint8_t hvx)
{
  uint32_t error_code;
  uint16_t i = 0;
  ble_gatts_value_t data_write = {data_size, 0, data};
  uint16_t u_connect_handle;
  u_connect_handle = m_connection_handle;
  error_code = sd_ble_gatts_sys_attr_set(m_connection_handle, NULL, 0, 0);
  error_code = sd_ble_gatts_value_set(u_connect_handle, handle, &data_write);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTS Value Set Error Code: 0x", error_code, 1);
    return error_code;
  }
  if (error_code == NRF_SUCCESS)
  {
    if (m_connection_handle != BLE_CONN_HANDLE_INVALID && hvx == 1)
    {
      ble_gatts_hvx_params_t m_hvx;
      m_hvx.handle = handle;
      m_hvx.offset = 0;
      m_hvx.p_data = data;
      m_hvx.p_len = &data_size;
      m_hvx.type = BLE_GATT_HVX_NOTIFICATION;
      error_code = sd_ble_gatts_hvx(m_connection_handle, &m_hvx);
      // //printf("sd_ble_gatts_hvx error code = 0x%02X\n", error_code);
      if (error_code != NRF_SUCCESS)
      {
        BLE_USB_PRINT_MSG_NUM("GATTS HVX Error Code: 0x", error_code, 1);
        return error_code;
      }
    }
  }
  BLE_USB_PRINT_MSG("GATTS Value Set OK\n");
  return error_code;
}

uint32_t gatts_char_value_get(uint16_t handle)
{
  uint32_t error_code = NRF_SUCCESS;
  ble_gatts_value_t r_value;
  error_code = sd_ble_gatts_value_get(m_connection_handle, handle, &r_value);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTS Value Get Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GATTS Value Get OK\n");
  return error_code;
}

uint32_t m_ble_stack_init()
{
  memset(m_connected_addr, 0, 6);
  uint32_t error_code;
  m_connection_handle = BLE_CONN_HANDLE_INVALID;
  error_code = ble_stack_init();
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("BLE Stack Init Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("BLE Stack Init OK\n");
  return error_code;
}

uint32_t scan_stop()
{
  uint32_t error_code;
  error_code = sd_ble_gap_scan_stop();
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GAP Scan Stop Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GAP Scan Stop OK\n");
  return error_code;
}

uint32_t b_connect(char *address, uint8_t addr_type, uint32_t timeout)
{
  ble_gap_addr_t *addr_gap;
  if (string_to_ble_address_convert(address, addr_gap) == false)
  {
    return 1;
  }
  addr_gap->addr_type = addr_type;
  addr_gap->addr_id_peer = 0;
  uint32_t error_code = 0;

  error_code = sd_ble_gap_connect(addr_gap, &m_scan_param, &m_connection_param, m_config_id);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GAP Connect Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GAP Connect Start\n");

  return error_code;
}

uint32_t descriptor_discover(ble_gattc_handle_range_t handle_range)
{
  uint32_t error_code = 0;
  uint16_t _1 = 1;

  uint16_t countx = (uint16_t)ceil(handle_range.end_handle - handle_range.start_handle) + _1;
  for (int x = 0; x < countx; x++)
  {
    ble_gattc_handle_range_t range = {(uint16_t)(handle_range.start_handle + (uint16_t)x), (uint16_t)handle_range.end_handle + _1};
    if ((error_code = sd_ble_gattc_descriptors_discover(m_connection_handle, &range)) != NRF_SUCCESS)
    {
      if (error_code != 0x11)
      {
        BLE_USB_PRINT_MSG_NUM("GATTC Descriptors Discover Error Code: 0x", error_code, 1);
        return error_code;
      }
    }
    if ((error_code = wait_trigger_evt(BLE_GATTC_EVT_DESC_DISC_RSP)) != NRF_SUCCESS)
    {
      BLE_USB_PRINT_MSG_NUM("GATTC Descriptors Discover Error Code: 0x", error_code, 1);
      return error_code;
    }
  }
  return error_code;
}

static uint32_t _128bit_uuid_char_discover(uint16_t handle)
{
  uint32_t error_code;
  ble_gattc_handle_range_t _handle_range;
  uint16_t i = 0;
  trigger_evt_id = 0;
  _handle_range.start_handle = _handle_range.end_handle = handle;
  discover_128_uuid = 1;
  error_code = sd_ble_gattc_attr_info_discover(m_connection_handle, &_handle_range);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Characteristic UUID Discover Error Code: 0x", error_code, 1);
    return error_code;
  }
  error_code = wait_trigger_evt(BLE_GATTC_EVT_ATTR_INFO_DISC_RSP);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Characteristic UUID Discover Error Code: 0x", error_code, 1);
    return error_code;
  }
  discover_128_uuid = 0;
  return NRF_SUCCESS;
}

uint32_t characteristic_discover(ble_gattc_handle_range_t handle_range)
{
  uint32_t current_size = 0;
  uint32_t error_code = 0;
  for (uint16_t i = 0; i < char_handle_index; i++)
  {
    trigger_evt_id = 0;

    // sd_ble_gattc_characteristics_discover(), it only discover 3 characteristics each time.
    uint16_t count = (uint16_t)((handle_range.end_handle - handle_range.start_handle - 1) / 3);
    for (uint16_t _index = 0; _index < count; _index++)
    {
      ble_gattc_handle_range_t range = {
          (uint16_t)(handle_range.start_handle + (1 + _index * 3)), (handle_range.start_handle + ((_index + 1) * 3))};
      trigger_evt_id = 0;
      error_code =
          sd_ble_gattc_characteristics_discover(m_connection_handle, &range);
      if (error_code != 0x11)
      {
        if (error_code != NRF_SUCCESS)
        {
          BLE_USB_PRINT_MSG_NUM("GATTC Characteristic Discover Error Code: 0x", error_code, 1);
          return error_code;
        }
        if ((error_code = wait_trigger_evt(BLE_GATTC_EVT_CHAR_DISC_RSP)) != NRF_SUCCESS)
        {
          BLE_USB_PRINT_MSG_NUM("GATTC Characteristic Discover Error Code: 0x", error_code, 1);
          return error_code;
        }
      }
      else
        error_code = 0;
    }
    // Get char's data
    for (uint16_t j = current_size; j < char_handle_index; j++)
    {
      uint16_t c_char_handle = gattc_char_handle[j];
      trigger_evt_id = 0;
      error_code = sd_ble_gattc_read(m_connection_handle, c_char_handle, 0);
      if (error_code != 0x11)
      {
        if (error_code != NRF_SUCCESS)
        {
          BLE_USB_PRINT_MSG_NUM("GATTC Characteristic Data Read Error Code: 0x", error_code, 1);
          return error_code;
        }
        else
        {
          error_code = wait_trigger_evt(BLE_GATTC_EVT_READ_RSP);
          if (error_code != NRF_SUCCESS)
          {
            BLE_USB_PRINT_MSG_NUM("GATTC Characteristic Data Read Error Code: 0x", error_code, 1);
            return error_code;
          }
        }
      }
    }
  }

  // Get char's 128 uuid
  for (uint16_t i = 0; i < _128_char_handle_index; i++)
  {
    trigger_evt_id = 0;
    // Get 128bit uuid
    // printf("Get Handle: 0x%02X's uuid\n", gattc_128_char_handle[i]);
    BLE_USB_PRINT_MSG_NUM("Get 128bit UUID Hanlde = 0x", gattc_128_char_handle[i], 1);
    ble_gattc_attr_info128_t _temp_info128 = {0, {0}};
    _temp_info128.handle = gattc_128_char_handle[i];
    error_code = _128bit_uuid_char_discover(gattc_128_char_handle[i]);
    if (error_code != 0x11)
    {
      if (error_code != NRF_SUCCESS)
        return error_code;
    }
  }

  for (uint16_t i = 0; i < _128_char_handle_index; i++)
  {
    uint16_t c_char_handle = gattc_128_char_handle[i];
    trigger_evt_id = 0;
    error_code = sd_ble_gattc_read(m_connection_handle, c_char_handle, 0);
    if (error_code != 0x11)
    {
      if (error_code != NRF_SUCCESS)
      {
        BLE_USB_PRINT_MSG_NUM("GATTC Characteristic Data Read Error Code: 0x", error_code, 1);
        return error_code;
      }
      else
      {
        error_code = wait_trigger_evt(BLE_GATTC_EVT_READ_RSP);
        if (error_code != NRF_SUCCESS)
        {
          BLE_USB_PRINT_MSG_NUM("GATTC Characteristic Data Read Error Code: 0x", error_code, 1);
          return error_code;
        }
      }
    }
  }

  error_code = descriptor_discover(handle_range);
  if (error_code != NRF_SUCCESS)
    return error_code;

  return error_code;
}

uint32_t b_client_discovery_start(uint8_t *uuid)
{
  uint32_t error_code = 0;

  memset(gattc_service_handle, 0, sizeof(ble_gattc_handle_range_t));
  memset(gattc_char_handle, 0, 50);
  memset(gattc_128_char_handle, 0, 50);
  memset(cccd_handle, 0, 100);
  memset(sccd_handle, 0, 100);

  service_handle_index = 0;
  char_handle_index = 0;
  _128_char_handle_index = 0;
  cccd_handle_index = 0;
  sccd_handle_index = 0;

  trigger_evt_id = 0;
  error_code = service_discovery_start(0x01, uuid);
  if (error_code != NRF_SUCCESS)
  {
    return error_code;
  }

  for (uint8_t i = 0; i < service_handle_index; i++)
  {
    error_code = characteristic_discover(gattc_service_handle[i]);
    if (error_code != NRF_SUCCESS)
    {
      return error_code;
    }
  }
  return 0;
}

uint32_t gattc_dissconect()
{
  uint32_t error_code;
  error_code = sd_ble_gap_disconnect(m_connection_handle,
                                     BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GAP Disconnect Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GAP Dissconnect OK\n");
  return error_code;
}

uint32_t get_addr()
{

  ble_gap_addr_t own_addr;
  uint32_t error_code;
  error_code = sd_ble_gap_addr_get(&own_addr);
  if (error_code != NRF_SUCCESS)
  {
    // printf("Error read address. Error Code = 0x%02X\n", error_code);
  }
  else
  {
    ble_usb_add_string("Address = ", 10);
    ble_usb_add_array(own_addr.addr, 6);
    ble_usb_add_string("\n", 1);
    ble_usb_printf();
    // printf("\n");
  }

  return error_code;
}

uint32_t update_mtu_size(uint16_t mtu_size)
{
  trigger_evt_id = 0;
  uint32_t error_code = 0;
  error_code = sd_ble_gattc_exchange_mtu_request(m_connection_handle, mtu_size);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC MTU Request Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GATTC MTU Request OK\n");
  return error_code;
}

uint32_t b_update_phy(uint8_t tx_phy, uint8_t rx_phy)
{
  uint32_t error_code = 0;
  ble_gap_phys_t phy_config = {tx_phy, rx_phy};
  trigger_evt_id = 0;
  error_code = sd_ble_gap_phy_update(m_connection_handle, &phy_config);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC PHY Update Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GATTC PHY Update OK\n");
  return error_code;
}

uint32_t enable_notification(uint16_t cccd_handle, uint16_t target_handle)
{
  uint8_t data_buffer[2] = {0x01, 0x00}; // 0x0001 = enable notify
  ble_gattc_write_params_t write_params;
  write_params.write_op = BLE_GATT_OP_WRITE_REQ;
  write_params.handle = cccd_handle;
  write_params.offset = 0;
  write_params.len = sizeof(data_buffer);
  write_params.p_value = data_buffer;

  uint32_t error_code = sd_ble_gattc_write(m_connection_handle, &write_params);
  if (error_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("GATTC Enable Notification Error Code: 0x", error_code, 1);
    return error_code;
  }
  BLE_USB_PRINT_MSG("GATTC Enable Notification OK\n");
  return error_code;
}

uint32_t get_cmd(char *rx_bfr, int len)
{
  char offset = 4;
  char *eq = strchr(rx_bfr + offset, '=');

  if (STR_CMP(rx_bfr + offset, "Connect", 7))
  {
    uint8_t addr_type;
    ble_gap_addr_t addr;
    sscanf(eq - 7, "Connect=%d_", &addr_type);
    string_to_ble_address_convert(eq + 3, &addr);
    return (b_connect(eq + 3, addr_type, 10));
  }
  else if (STR_CMP(rx_bfr + offset, "Init", 4))
  {
    m_ble_stack_init();
  }
  else if (STR_CMP(rx_bfr + offset, "Disconnect", 10))
  {
    gattc_dissconect();
  }
  else if (STR_CMP(rx_bfr + offset, "Scan_Strat", 10))
  {
    scan_start();
  }
  else if (STR_CMP(rx_bfr + offset, "Scan_Stop", 9))
  {
    scan_stop();
  }
  else if (STR_CMP(rx_bfr + offset, "Discovery", 9))
  {
    b_client_discovery_start(NULL);
  }
  else if (STR_CMP(rx_bfr + offset, "Get_Addr", 8))
  {
    get_addr();
  }
  else if (STR_CMP(rx_bfr + offset, "Update_MTU", 10))
  {
    uint16_t mtu_size;
    sscanf(eq + 1, "=%d", &mtu_size);
    update_mtu_size(mtu_size);
  }
  else if (STR_CMP(rx_bfr + offset, "Update_PHY", 10))
  {
    uint8_t tx_phy, rx_phy;
    sscanf(eq + 1, "=%d_%d", &tx_phy, &rx_phy);
    b_update_phy(tx_phy, rx_phy);
  }
  else if (STR_CMP(rx_bfr + offset, "Enable_Notify", 13))
  {
    uint16_t cccd_handle, target_handle;
    sscanf(eq + 1, "=%d_%d", &cccd_handle, &target_handle);
    enable_notification(cccd_handle, target_handle);
  }
  return 0;
}