#ifndef MEGASIG_BLE_FUNCTION_H
#define MEGASIG_BLE_FUNCTION_H

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <string.h>
#include <stdarg.h>

#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "ble.h"
#include "ble_hci.h"
#include "app_timer.h"
#include "nrf_delay.h"
#include "my_uart.h"

#define APP_TIMER_TICKS_TO_MS(TICK) ((TICK * 1000 * (APP_TIMER_CONFIG_RTC_FREQUENCY + 1)) / APP_TIMER_CLOCK_FREQ)

#define DATA_T(str) ((data_t){(const uint8_t *)(str), (uint16_t)strlen(str)})

#define BLE_EVT_BUF_SIZE 2048

static uint8_t m_ble_evt_buf[BLE_EVT_BUF_SIZE];

void ble_usb_printf(void);
void ble_usb_add_array(const uint8_t *value, uint32_t len);
void ble_usb_add_value_uint8(uint8_t value, uint8_t type);
void ble_usb_add_value_uint16(uint16_t value, uint8_t type);
void ble_usb_add_value_uint32(uint32_t value, uint8_t type);
void ble_usb_add_string(uint8_t *string, uint16_t len);
void ble_usb_add_char_props(ble_gatt_char_props_t props);

#define BLE_USB_PRINT_MSG_NUM(msg, error_code, type) \
  do                                                 \
  {                                                  \
    ble_usb_add_string((msg), strlen(msg) - 1);      \
    ble_usb_add_value_uint32((error_code), (type));  \
    ble_usb_add_string("\n", 1);                     \
    ble_usb_printf();                                \
  } while (0)

#define BLE_USB_PRINT_MSG(msg)                  \
  do                                            \
  {                                             \
    ble_usb_add_string((msg), strlen(msg) - 1); \
    ble_usb_printf();                           \
  } while (0)

#define STR_CMP(start, target, len) strncmp(start, target, len) == 0

void ble_address_to_string_convert(ble_gap_addr_t address, uint8_t *string_buffer);
bool find_adv_name(const ble_gap_evt_adv_report_t *p_adv_report, const char *name_to_find);
void ble_event_get_with_exe(void);
uint32_t scan_start();
static uint32_t gattc_read(uint16_t target_handle);
uint32_t _128bit_uuid_service_discover();
uint32_t service_discovery_start(uint16_t start_handle, uint8_t *uuid);
uint32_t gattc_write(uint16_t handle, uint8_t write_op, uint32_t data_len, uint8_t *data);
uint32_t gattc_request_mtu_size(uint8_t mtu_size);
uint32_t gattc_services_discover(uint16_t start_handle, uint16_t uuid, uint8_t type);
uint32_t gattc_characteristics_discover(uint16_t start_handle, uint16_t end_handle);
uint32_t gattc_descriptors_discover(uint16_t start_handle, uint16_t end_handle);
uint32_t gattc_attr_info_discover(uint16_t start_handle, uint16_t end_handle);
uint32_t b_characteristic_init(uint16_t service_handle, uint16_t char_uuid, uint8_t *char_name,
                               uint8_t char_name_size, uint8_t *char_props, uint8_t read_perm,
                               uint8_t write_perm, uint8_t rd_auth, uint8_t wr_auth,
                               uint8_t ex_reliable_wr, uint8_t ex_wr_aux, uint16_t *handle,
                               uint16_t *cccd_handle, uint16_t *sccd_handle,
                               uint16_t *user_desc_handle);
uint32_t b_services_init(uint16_t u_uuid);
uint32_t advertising_init(uint8_t adv_AD_type, uint8_t *adv_data_data,
                          uint8_t adv_data_len, uint16_t interval, uint16_t duration);
uint32_t advertising_start(uint16_t *adv_handle, uint16_t count);
uint32_t advertising_stop(uint16_t adv_handle);
uint32_t gatts_char_value_set(uint16_t handle, uint8_t *data, uint16_t data_size, uint8_t hvx);
uint32_t gatts_char_value_get(uint16_t handle);
// uint32_t m_ble_cfg_set();
uint32_t m_ble_stack_init();
uint32_t scan_stop();
uint32_t connect(char *address, uint8_t addr_type, uint32_t timeout);
uint32_t descriptor_discover(ble_gattc_handle_range_t handle_range);
static uint32_t _128bit_uuid_char_discover(uint16_t handle);
uint32_t characteristic_discover(ble_gattc_handle_range_t handle_range);
uint32_t b_client_discovery_start(uint8_t *uuid);
uint32_t gattc_dissconect();
uint32_t get_addr();
uint32_t update_mtu_size(uint16_t mtu_size);
uint32_t b_update_phy(uint8_t tx_phy, uint8_t rx_phy);
uint32_t enable_notification(uint16_t cccd_handle, uint16_t target_handle);
uint32_t get_cmd(char *rx_bfr, int len);
#endif