#ifndef MEGASIG_GLOBAL_H
#define MEGASIG_GLOBAL_H

/** Definitions */
#define DEFAULT_BAUD_RATE \
  2000000 /**< The baud rate to be used for serial communication with nRF5 device. */

#define MSEC_TO_UNITS(TIME, RESOLUTION) (((TIME) * 1000) / (RESOLUTION))

#define SCAN_INTERVAL 0x00A0 /**< Determines scan interval in units of 0.625 milliseconds. */
#define SCAN_WINDOW 0x0050   /**< Determines scan window in units of 0.625 milliseconds. */
#define SCAN_TIMEOUT \
  0x0 /**< Scan timeout between 0x01 and 0xFFFF in seconds, 0x0 disables timeout. */

#define MIN_CONNECTION_INTERVAL                                                                 \
  MSEC_TO_UNITS(7.5, UNIT_1_25_MS) /**< Determines minimum connection interval in milliseconds. \
                                    */
#define MAX_CONNECTION_INTERVAL                                                                \
  MSEC_TO_UNITS(30, UNIT_1_25_MS) /**< Determines maximum connection interval in milliseconds. \
                                   */
#define SLAVE_LATENCY 0           /**< Slave Latency in number of connection events. */
#define CONNECTION_SUPERVISION_TIMEOUT \
  MSEC_TO_UNITS(4000,                  \
                UNIT_10_MS) /**< Determines supervision time-out in units of 10 milliseconds. */

#define TARGET_DEV_NAME \
  "Nordic_UART"          /**< Connect to a peripheral using a given advertising name here. */
#define MAX_PEER_COUNT 1 /**< Maximum number of peer's application intends to manage. */

#define BLE_UUID_NUS_SERVICE 0x0001           /**< The UUID of the Nordic UART Service. */
#define BLE_UUID_NUS_RX_CHARACTERISTIC 0x0002 /**< The UUID of the RX Characteristic. */
#define BLE_UUID_NUS_TX_CHARACTERISTIC 0x0003 /**< The UUID of the TX Characteristic. */

#define NUS_BASE_UUID                                                                     \
  {                                                                                       \
    {                                                                                     \
      0x9E, 0xCA, 0xDC, 0x24, 0x0E, 0xE5, 0xA9, 0xE0, 0x93, 0xF3, 0xA3, 0xB5, 0x00, 0x00, \
          0x40, 0x6E                                                                      \
    }                                                                                     \
  } /**< Used vendor specific UUID. */

#define BLE_UUID_CCCD 0x2902
#define BLE_UUID_CHARACTERISTIC 0x2803
#define BLE_CCCD_NOTIFY 0x01

#define SRV_DISC_START_HANDLE 0x0001 /**< The start handle value used during service discovery. */

#define BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND \
  0x010A /**< ATT Error: Used in ATT as Attribute not found. */

#define STRING_BUFFER_SIZE 50

#define PRIVATE_KEY_SIZE 32
#define PUBLIC_KEY_SIZE 64
#define DH_KEY_SIZE 32

#define BLE_MAX_BFR_SIZE 1024

#define BLE_EVT_BUF_SIZE 2048

static __align(BLE_EVT_PTR_ALIGNMENT) uint8_t m_ble_evt_buf[BLE_EVT_BUF_SIZE];

typedef struct
{
  uint8_t *p_data;   /**< Pointer to data. */
  uint16_t data_len; /**< Length of data. */
} data_t;

typedef struct
{
  uint16_t handle;
  uint8_t data[251];
} hvx_data;

static uint8_t m_connected_devices = 0;
static uint16_t m_connection_handle = BLE_CONN_HANDLE_INVALID;
static uint8_t m_connected_addr[6] = {0};
static uint16_t m_start_handle = 0x01;
static uint16_t m_end_handle = 0xffff;
static uint16_t m_nus_rx_value_handle = 0;
static uint16_t m_nus_tx_value_handle = 0;
static uint16_t m_nus_cccd_handle = 0;
static bool m_connection_is_in_progress = false;
static bool connect_status = false;
static bool act_status = false;
static ble_gap_sec_params_t m_sec_params;
static uint8_t passkey[6] = {0};
static ble_gap_evt_adv_report_t current_scan_info;

static ble_gattc_handle_range_t gattc_service_handle[20];
static uint8_t service_handle_index = 0;
static ble_gattc_attr_info128_t gattc_128_service_handle[20];
static uint8_t service_128_handle_index = 0;
static uint16_t gattc_char_handle[50];
static uint8_t char_handle_index = 0;
static uint16_t gattc_128_char_handle[50];
static uint8_t _128_char_handle_index = 0;
static uint16_t cccd_handle[100];
static uint8_t cccd_handle_index = 0;
static uint16_t sccd_handle[100];
static uint8_t sccd_handle_index = 0;

static uint32_t m_config_id = 1;
static uint8_t mp_data[100] = {0};
static ble_data_t m_adv_report_buffer;

static uint8_t op_bfr[BLE_MAX_BFR_SIZE];
static uint16_t op_bfr_count;

static uint8_t discover_128_uuid = 0;

static uint16_t trigger_evt_id;

static const ble_gap_scan_params_t m_scan_param = {
    0, // Set if accept extended advertising packetets.
    0, // Set if report inomplete reports.
    0, // Set if active scanning.
    BLE_GAP_SCAN_FP_ACCEPT_ALL,
    BLE_GAP_PHY_1MBPS,
    (uint16_t)SCAN_INTERVAL,
    (uint16_t)SCAN_WINDOW,
    (uint16_t)SCAN_TIMEOUT,
    {0} // Set chennel mask.
};

static const ble_gap_conn_params_t m_connection_param = {
    (uint16_t)MIN_CONNECTION_INTERVAL, (uint16_t)MAX_CONNECTION_INTERVAL, (uint16_t)SLAVE_LATENCY,
    (uint16_t)CONNECTION_SUPERVISION_TIMEOUT};

static ble_data_t m_adv_report_data = {mp_data, 100};

#endif // MEGASIG_GLOBAL_H
